from django.test import TestCase

from haystack import connections, inputs


class Elasticsearch5InputTestCase(TestCase):
    def setUp(self):
        super().setUp()
        self.query_obj = connections["elasticsearch"].get_query()

    def test_raw_init(self):
        raw = inputs.Raw("hello OR there, :you")
        self.assertEqual(raw.query_string, "hello OR there, :you")
        self.assertEqual(raw.kwargs, {})
        self.assertEqual(raw.post_process, False)

        raw = inputs.Raw("hello OR there, :you", test="really")
        self.assertEqual(raw.query_string, "hello OR there, :you")
        self.assertEqual(raw.kwargs, {"test": "really"})
        self.assertEqual(raw.post_process, False)

    def test_raw_prepare(self):
        raw = inputs.Raw("hello OR there, :you")
        self.assertEqual(raw.prepare(self.query_obj), "hello OR there, :you")

    def test_clean_init(self):
        clean = inputs.Clean("hello OR there, :you")
        self.assertEqual(clean.query_string, "hello OR there, :you")
        self.assertEqual(clean.post_process, True)

    def test_clean_prepare(self):
        clean = inputs.Clean("hello OR there, :you")
        self.assertEqual(clean.prepare(self.query_obj), "hello or there, \\:you")

    def test_exact_init(self):
        exact = inputs.Exact("hello OR there, :you")
        self.assertEqual(exact.query_string, "hello OR there, :you")
        self.assertEqual(exact.post_process, True)

    def test_exact_prepare(self):
        exact = inputs.Exact("hello OR there, :you")
        self.assertEqual(exact.prepare(self.query_obj), '"hello OR there, :you"')

        exact = inputs.Exact("hello OR there, :you", clean=True)
        self.assertEqual(exact.prepare(self.query_obj), '"hello or there, \\:you"')

    def test_not_init(self):
        not_it = inputs.Not("hello OR there, :you")
        self.assertEqual(not_it.query_string, "hello OR there, :you")
        self.assertEqual(not_it.post_process, True)

    def test_not_prepare(self):
        not_it = inputs.Not("hello OR there, :you")
        self.assertEqual(not_it.prepare(self.query_obj), "NOT (hello or there, \\:you)")

    def test_autoquery_init(self):
        autoquery = inputs.AutoQuery('panic -don\'t "froody dude"')
        self.assertEqual(autoquery.query_string, 'panic -don\'t "froody dude"')
        self.assertEqual(autoquery.post_process, False)

    def test_autoquery_prepare(self):
        autoquery = inputs.AutoQuery('panic -don\'t "froody dude"')
        self.assertEqual(
            autoquery.prepare(self.query_obj), 'panic NOT don\'t "froody dude"'
        )

    def test_altparser_init(self):
        altparser = inputs.AltParser("dismax")
        self.assertEqual(altparser.parser_name, "dismax")
        self.assertEqual(altparser.query_string, "")
        self.assertEqual(altparser.kwargs, {})
        self.assertEqual(altparser.post_process, False)

        altparser = inputs.AltParser("dismax", "douglas adams", qf="author", mm=1)
        self.assertEqual(altparser.parser_name, "dismax")
        self.assertEqual(altparser.query_string, "douglas adams")
        self.assertEqual(altparser.kwargs, {"mm": 1, "qf": "author"})
        self.assertEqual(altparser.post_process, False)

    def test_altparser_prepare(self):
        altparser = inputs.AltParser("dismax", "douglas adams", qf="author", mm=1)
        self.assertEqual(
            altparser.prepare(self.query_obj),
            """{!dismax mm=1 qf=author v='douglas adams'}""",
        )



from django.test import TestCase

from haystack import connection_router, connections
from haystack.forms import FacetedSearchForm, ModelSearchForm, SearchForm, model_choices
from haystack.query import EmptySearchQuerySet, SearchQuerySet
from haystack.utils.loading import UnifiedIndex
from test_haystack.core.models import AnotherMockModel, MockModel
from test_haystack.test_views import (
    BasicAnotherMockModelSearchIndex,
    BasicMockModelSearchIndex,
)


class SearchFormTestCase(TestCase):
    def setUp(self):
        super().setUp()

        # Stow.
        self.old_unified_index = connections["default"]._index
        self.ui = UnifiedIndex()
        self.bmmsi = BasicMockModelSearchIndex()
        self.bammsi = BasicAnotherMockModelSearchIndex()
        self.ui.build(indexes=[self.bmmsi, self.bammsi])
        connections["default"]._index = self.ui

        # Update the "index".
        backend = connections["default"].get_backend()
        backend.clear()
        backend.update(self.bmmsi, MockModel.objects.all())

        self.sqs = SearchQuerySet()

    def tearDown(self):
        connections["default"]._index = self.old_unified_index
        super().tearDown()

    def test_unbound(self):
        sf = SearchForm({}, searchqueryset=self.sqs)

        self.assertEqual(sf.errors, {})
        self.assertEqual(sf.is_valid(), True)

        # This shouldn't blow up.
        sqs = sf.search()
        self.assertTrue(isinstance(sqs, EmptySearchQuerySet))


class ModelSearchFormTestCase(TestCase):
    def setUp(self):
        super().setUp()
        # Stow.
        self.old_unified_index = connections["default"]._index
        self.ui = UnifiedIndex()
        self.bmmsi = BasicMockModelSearchIndex()
        self.bammsi = BasicAnotherMockModelSearchIndex()
        self.ui.build(indexes=[self.bmmsi, self.bammsi])
        connections["default"]._index = self.ui

        # Update the "index".
        backend = connections["default"].get_backend()
        backend.clear()
        backend.update(self.bmmsi, MockModel.objects.all())

        self.sqs = SearchQuerySet()

    def tearDown(self):
        connections["default"]._index = self.old_unified_index
        super().tearDown()

    def test_models_regression_1(self):
        # Regression for issue #1.
        msf = ModelSearchForm(
            {"query": "test", "models": ["core.mockmodel", "core.anothermockmodel"]},
            searchqueryset=self.sqs,
        )

        self.assertEqual(
            msf.fields["models"].choices,
            [
                ("core.anothermockmodel", "Another mock models"),
                ("core.mockmodel", "Mock models"),
            ],
        )
        self.assertEqual(msf.errors, {})
        self.assertEqual(msf.is_valid(), True)

        sqs_with_models = msf.search()
        self.assertEqual(len(sqs_with_models.query.models), 2)

    def test_model_choices(self):
        self.assertEqual(len(model_choices()), 2)
        self.assertEqual(
            [option[1] for option in model_choices()],
            ["Another mock models", "Mock models"],
        )

    def test_model_choices_unicode(self):
        stowed_verbose_name_plural = MockModel._meta.verbose_name_plural
        MockModel._meta.verbose_name_plural = "☃"
        self.assertEqual(len(model_choices()), 2)
        self.assertEqual(
            [option[1] for option in model_choices()], ["Another mock models", "☃"]
        )
        MockModel._meta.verbose_name_plural = stowed_verbose_name_plural


class FacetedSearchFormTestCase(TestCase):
    def setUp(self):
        super().setUp()
        # Stow.
        self.old_unified_index = connections["default"]._index
        self.ui = UnifiedIndex()
        self.bmmsi = BasicMockModelSearchIndex()
        self.bammsi = BasicAnotherMockModelSearchIndex()
        self.ui.build(indexes=[self.bmmsi, self.bammsi])
        connections["default"]._index = self.ui

        # Update the "index".
        backend = connections["default"].get_backend()
        backend.clear()
        backend.update(self.bmmsi, MockModel.objects.all())

        self.sqs = SearchQuerySet()

    def tearDown(self):
        connections["default"]._index = self.old_unified_index
        super().tearDown()

    def test_init_with_selected_facets(self):
        sf = FacetedSearchForm({}, searchqueryset=self.sqs)
        self.assertEqual(sf.errors, {})
        self.assertEqual(sf.is_valid(), True)
        self.assertEqual(sf.selected_facets, [])

        sf = FacetedSearchForm({}, selected_facets=[], searchqueryset=self.sqs)
        self.assertEqual(sf.errors, {})
        self.assertEqual(sf.is_valid(), True)
        self.assertEqual(sf.selected_facets, [])

        sf = FacetedSearchForm(
            {}, selected_facets=["author:daniel"], searchqueryset=self.sqs
        )
        self.assertEqual(sf.errors, {})
        self.assertEqual(sf.is_valid(), True)
        self.assertEqual(sf.selected_facets, ["author:daniel"])

        sf = FacetedSearchForm(
            {},
            selected_facets=["author:daniel", "author:chris"],
            searchqueryset=self.sqs,
        )
        self.assertEqual(sf.errors, {})
        self.assertEqual(sf.is_valid(), True)
        self.assertEqual(sf.selected_facets, ["author:daniel", "author:chris"])

    def test_search(self):
        sf = FacetedSearchForm(
            {"q": "test"}, selected_facets=[], searchqueryset=self.sqs
        )
        sqs = sf.search()
        self.assertEqual(sqs.query.narrow_queries, set())

        # Test the "skip no-colon" bits.
        sf = FacetedSearchForm(
            {"q": "test"}, selected_facets=["authordaniel"], searchqueryset=self.sqs
        )
        sqs = sf.search()
        self.assertEqual(sqs.query.narrow_queries, set())

        sf = FacetedSearchForm(
            {"q": "test"}, selected_facets=["author:daniel"], searchqueryset=self.sqs
        )
        sqs = sf.search()
        self.assertEqual(sqs.query.narrow_queries, set(['author:"daniel"']))

        sf = FacetedSearchForm(
            {"q": "test"},
            selected_facets=["author:daniel", "author:chris"],
            searchqueryset=self.sqs,
        )
        sqs = sf.search()
        self.assertEqual(
            sqs.query.narrow_queries, set(['author:"daniel"', 'author:"chris"'])
        )



import os
import unittest

from django.conf import settings

from haystack.utils import log as logging


def load_tests(loader, standard_tests, pattern):
    log = logging.getLogger("haystack")
    try:
        import elasticsearch

        if not ((1, 0, 0) <= elasticsearch.__version__ < (2, 0, 0)):
            raise ImportError
        from elasticsearch import Elasticsearch, ElasticsearchException
    except ImportError:
        log.error(
            "Skipping ElasticSearch 1 tests: 'elasticsearch>=1.0.0,<2.0.0' not installed."
        )
        raise unittest.SkipTest("'elasticsearch>=1.0.0,<2.0.0' not installed.")

    es = Elasticsearch(settings.HAYSTACK_CONNECTIONS["elasticsearch"]["URL"])
    try:
        es.info()
    except ElasticsearchException as e:
        log.error(
            "elasticsearch not running on %r"
            % settings.HAYSTACK_CONNECTIONS["elasticsearch"]["URL"],
            exc_info=True,
        )
        raise unittest.SkipTest(
            "elasticsearch not running on %r"
            % settings.HAYSTACK_CONNECTIONS["elasticsearch"]["URL"],
            e,
        )

    package_tests = loader.discover(
        start_dir=os.path.dirname(__file__), pattern=pattern
    )
    standard_tests.addTests(package_tests)
    return standard_tests



#!/usr/bin/env python
import os
import sys

import django
from django.core.management import call_command


def run_all(argv=None):
    sys.path.insert(0, os.path.dirname(os.path.dirname(__file__)))
    os.environ["DJANGO_SETTINGS_MODULE"] = "test_haystack.settings"
    django.setup()

    call_command("test", sys.argv[1:])


if __name__ == "__main__":
    run_all(sys.argv)



import queue
import time
from threading import Thread

from django import forms
from django.http import HttpRequest, QueryDict
from django.test import TestCase, override_settings
from django.urls import reverse

from haystack import connections, indexes
from haystack.forms import FacetedSearchForm, ModelSearchForm, SearchForm
from haystack.query import EmptySearchQuerySet
from haystack.utils.loading import UnifiedIndex
from haystack.views import FacetedSearchView, SearchView, search_view_factory
from test_haystack.core.models import AnotherMockModel, MockModel


class InitialedSearchForm(SearchForm):
    q = forms.CharField(initial="Search for...", required=False, label="Search")


class BasicMockModelSearchIndex(indexes.BasicSearchIndex, indexes.Indexable):
    def get_model(self):
        return MockModel


class BasicAnotherMockModelSearchIndex(indexes.BasicSearchIndex, indexes.Indexable):
    def get_model(self):
        return AnotherMockModel


class SearchViewTestCase(TestCase):
    fixtures = ["base_data"]

    def setUp(self):
        super().setUp()

        # Stow.
        self.old_unified_index = connections["default"]._index
        self.ui = UnifiedIndex()
        self.bmmsi = BasicMockModelSearchIndex()
        self.bammsi = BasicAnotherMockModelSearchIndex()
        self.ui.build(indexes=[self.bmmsi, self.bammsi])
        connections["default"]._index = self.ui

        # Update the "index".
        backend = connections["default"].get_backend()
        backend.clear()
        backend.update(self.bmmsi, MockModel.objects.all())

    def tearDown(self):
        connections["default"]._index = self.old_unified_index
        super().tearDown()

    def test_search_no_query(self):
        response = self.client.get(reverse("haystack_search"))
        self.assertEqual(response.status_code, 200)

    def test_search_query(self):
        response = self.client.get(reverse("haystack_search"), {"q": "haystack"})
        self.assertEqual(response.status_code, 200)
        self.assertIn("page", response.context)
        self.assertNotIn("page_obj", response.context)
        self.assertEqual(len(response.context[-1]["page"].object_list), 3)
        self.assertEqual(
            response.context[-1]["page"].object_list[0].content_type(), "core.mockmodel"
        )
        self.assertEqual(response.context[-1]["page"].object_list[0].pk, "1")

    def test_invalid_page(self):
        response = self.client.get(
            reverse("haystack_search"), {"q": "haystack", "page": "165233"}
        )
        self.assertEqual(response.status_code, 404)

    def test_empty_results(self):
        sv = SearchView()
        sv.request = HttpRequest()
        sv.form = sv.build_form()
        self.assertTrue(isinstance(sv.get_results(), EmptySearchQuerySet))

    def test_initial_data(self):
        sv = SearchView(form_class=InitialedSearchForm)
        sv.request = HttpRequest()
        form = sv.build_form()
        self.assertTrue(isinstance(form, InitialedSearchForm))
        self.assertEqual(form.fields["q"].initial, "Search for...")
        para = form.as_p()
        self.assertTrue('<label for="id_q">Search:</label>' in para)
        self.assertTrue('value="Search for..."' in para)

    def test_pagination(self):
        response = self.client.get(
            reverse("haystack_search"), {"q": "haystack", "page": 0}
        )
        self.assertEqual(response.status_code, 404)
        response = self.client.get(
            reverse("haystack_search"), {"q": "haystack", "page": 1}
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.context[-1]["page"].object_list), 3)
        response = self.client.get(
            reverse("haystack_search"), {"q": "haystack", "page": 2}
        )
        self.assertEqual(response.status_code, 404)

    def test_thread_safety(self):
        exceptions = []

        def threaded_view(resp_queue, view, request):
            time.sleep(2)

            try:
                view(request)
                resp_queue.put(request.GET["name"])
            except Exception as e:
                exceptions.append(e)
                raise

        class ThreadedSearchView(SearchView):
            def __call__(self, request):
                print("Name: %s" % request.GET["name"])
                return super().__call__(request)

        view = search_view_factory(view_class=ThreadedSearchView)
        resp_queue = queue.Queue()
        request_1 = HttpRequest()
        request_1.GET = {"name": "foo"}
        request_2 = HttpRequest()
        request_2.GET = {"name": "bar"}

        th1 = Thread(target=threaded_view, args=(resp_queue, view, request_1))
        th2 = Thread(target=threaded_view, args=(resp_queue, view, request_2))

        th1.start()
        th2.start()
        th1.join()
        th2.join()

        foo = resp_queue.get()
        bar = resp_queue.get()
        self.assertNotEqual(foo, bar)

    def test_spelling(self):
        # Stow.
        from django.conf import settings

        old = settings.HAYSTACK_CONNECTIONS["default"].get("INCLUDE_SPELLING", None)

        settings.HAYSTACK_CONNECTIONS["default"]["INCLUDE_SPELLING"] = True

        sv = SearchView()
        sv.query = "Nothing"
        sv.results = []
        sv.build_page = lambda: (None, None)
        sv.create_response()
        context = sv.get_context()

        self.assertIn(
            "suggestion",
            context,
            msg="Spelling suggestions should be present even if"
            " no results were returned",
        )
        self.assertEqual(context["suggestion"], None)

        # Restore
        settings.HAYSTACK_CONNECTIONS["default"]["INCLUDE_SPELLING"] = old

        if old is None:
            del settings.HAYSTACK_CONNECTIONS["default"]["INCLUDE_SPELLING"]


@override_settings(ROOT_URLCONF="test_haystack.results_per_page_urls")
class ResultsPerPageTestCase(TestCase):
    fixtures = ["base_data"]

    def setUp(self):
        super().setUp()

        # Stow.
        self.old_unified_index = connections["default"]._index
        self.ui = UnifiedIndex()
        self.bmmsi = BasicMockModelSearchIndex()
        self.bammsi = BasicAnotherMockModelSearchIndex()
        self.ui.build(indexes=[self.bmmsi, self.bammsi])
        connections["default"]._index = self.ui

        # Update the "index".
        backend = connections["default"].get_backend()
        backend.clear()
        backend.update(self.bmmsi, MockModel.objects.all())

    def tearDown(self):
        connections["default"]._index = self.old_unified_index
        super().tearDown()

    def test_custom_results_per_page(self):
        response = self.client.get("/search/", {"q": "haystack"})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.context[-1]["page"].object_list), 1)
        self.assertEqual(response.context[-1]["paginator"].per_page, 1)

        response = self.client.get("/search2/", {"q": "hello world"})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.context[-1]["page"].object_list), 2)
        self.assertEqual(response.context[-1]["paginator"].per_page, 2)


class FacetedSearchViewTestCase(TestCase):
    def setUp(self):
        super().setUp()

        # Stow.
        self.old_unified_index = connections["default"]._index
        self.ui = UnifiedIndex()
        self.bmmsi = BasicMockModelSearchIndex()
        self.bammsi = BasicAnotherMockModelSearchIndex()
        self.ui.build(indexes=[self.bmmsi, self.bammsi])
        connections["default"]._index = self.ui

        # Update the "index".
        backend = connections["default"].get_backend()
        backend.clear()
        backend.update(self.bmmsi, MockModel.objects.all())

    def tearDown(self):
        connections["default"]._index = self.old_unified_index
        super().tearDown()

    def test_search_no_query(self):
        response = self.client.get(reverse("haystack_faceted_search"))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.context["facets"], {})

    def test_empty_results(self):
        fsv = FacetedSearchView()
        fsv.request = HttpRequest()
        fsv.request.GET = QueryDict("")
        fsv.form = fsv.build_form()
        self.assertTrue(isinstance(fsv.get_results(), EmptySearchQuerySet))

    def test_default_form(self):
        fsv = FacetedSearchView()
        fsv.request = HttpRequest()
        fsv.request.GET = QueryDict("")
        fsv.form = fsv.build_form()
        self.assertTrue(isinstance(fsv.form, FacetedSearchForm))

    def test_list_selected_facets(self):
        fsv = FacetedSearchView()
        fsv.request = HttpRequest()
        fsv.request.GET = QueryDict("")
        fsv.form = fsv.build_form()
        self.assertEqual(fsv.form.selected_facets, [])

        fsv = FacetedSearchView()
        fsv.request = HttpRequest()
        fsv.request.GET = QueryDict(
            "selected_facets=author:daniel&selected_facets=author:chris"
        )
        fsv.form = fsv.build_form()
        self.assertEqual(fsv.form.selected_facets, ["author:daniel", "author:chris"])


class BasicSearchViewTestCase(TestCase):
    fixtures = ["base_data"]

    def setUp(self):
        super().setUp()

        # Stow.
        self.old_unified_index = connections["default"]._index
        self.ui = UnifiedIndex()
        self.bmmsi = BasicMockModelSearchIndex()
        self.bammsi = BasicAnotherMockModelSearchIndex()
        self.ui.build(indexes=[self.bmmsi, self.bammsi])
        connections["default"]._index = self.ui

        # Update the "index".
        backend = connections["default"].get_backend()
        backend.clear()
        backend.update(self.bmmsi, MockModel.objects.all())

    def tearDown(self):
        connections["default"]._index = self.old_unified_index
        super().tearDown()

    def test_search_no_query(self):
        response = self.client.get(reverse("haystack_basic_search"))
        self.assertEqual(response.status_code, 200)

    def test_search_query(self):
        response = self.client.get(reverse("haystack_basic_search"), {"q": "haystack"})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(type(response.context[-1]["form"]), ModelSearchForm)
        self.assertEqual(len(response.context[-1]["page"].object_list), 3)
        self.assertEqual(
            response.context[-1]["page"].object_list[0].content_type(), "core.mockmodel"
        )
        self.assertEqual(response.context[-1]["page"].object_list[0].pk, "1")
        self.assertEqual(response.context[-1]["query"], "haystack")

    def test_invalid_page(self):
        response = self.client.get(
            reverse("haystack_basic_search"), {"q": "haystack", "page": "165233"}
        )
        self.assertEqual(response.status_code, 404)



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.template import Context, Template
from django.test import TestCase

from haystack.utils.highlighting import Highlighter


class BorkHighlighter(Highlighter):
    def render_html(self, highlight_locations=None, start_offset=None, end_offset=None):
        highlighted_chunk = self.text_block[start_offset:end_offset]

        for word in self.query_words:
            highlighted_chunk = highlighted_chunk.replace(word, "Bork!")

        return highlighted_chunk


class TemplateTagTestCase(TestCase):
    def render(self, template, context):
        # Why on Earth does Django not have a TemplateTestCase yet?
        t = Template(template)
        c = Context(context)
        return t.render(c)


class HighlightTestCase(TemplateTagTestCase):
    def setUp(self):
        super().setUp()
        self.sample_entry = """
Registering indexes in Haystack is very similar to registering models and
ModelAdmin classes in the Django admin site. If you want to override the default
indexing behavior for your model you can specify your own SearchIndex class.
This is useful for ensuring that future-dated or non-live content is not indexed
and searchable.

Every custom SearchIndex requires there be one and only one field with
document=True. This is the primary field that will get passed to the backend
for indexing. For this field, you'll then need to create a template at
search/indexes/myapp/note_text.txt. This allows you to customize the document
that will be passed to the search backend for indexing. A sample template might
look like.

In addition, you may specify other fields to be populated along with the
document. In this case, we also index the user who authored the document as
well as the date the document was published. The variable you assign the
SearchField to should directly map to the field your search backend is
expecting. You instantiate most search fields with a parameter that points to
the attribute of the object to populate that field with.
"""

    def test_simple(self):
        template = """{% load highlight %}{% highlight entry with query %}"""
        context = {"entry": self.sample_entry, "query": "index"}
        self.assertEqual(
            self.render(template, context),
            '...<span class="highlighted">index</span>ing behavior for your model you can specify your own Search<span class="highlighted">Index</span> class.\nThis is useful for ensuring that future-dated or non-live content is not <span class="highlighted">index</span>ed\nand searchable.\n\nEvery custom Search<span class="highlighted">Index</span> ...',
        )

        template = """{% load highlight %}{% highlight entry with query html_tag "div" css_class "foo" max_length 100 %}"""
        context = {"entry": self.sample_entry, "query": "field"}
        self.assertEqual(
            self.render(template, context),
            '...<div class="foo">field</div> with\ndocument=True. This is the primary <div class="foo">field</div> that will get passed to the backend\nfor indexing...',
        )

        template = """{% load highlight %}{% highlight entry with query html_tag "div" css_class "foo" max_length 100 %}"""
        context = {"entry": self.sample_entry, "query": "Haystack"}
        self.assertEqual(
            self.render(template, context),
            '...<div class="foo">Haystack</div> is very similar to registering models and\nModelAdmin classes in the Django admin site. If y...',
        )

        template = """{% load highlight %}{% highlight "xxxxxxxxxxxxx foo bbxxxxx foo" with "foo" max_length 5 html_tag "span" %}"""
        context = {}
        self.assertEqual(
            self.render(template, context),
            '...<span class="highlighted">foo</span> b...',
        )

    def test_custom(self):
        # Stow.
        old_custom_highlighter = getattr(settings, "HAYSTACK_CUSTOM_HIGHLIGHTER", None)
        settings.HAYSTACK_CUSTOM_HIGHLIGHTER = "not.here.FooHighlighter"

        template = """{% load highlight %}{% highlight entry with query %}"""
        context = {"entry": self.sample_entry, "query": "index"}
        self.assertRaises(ImproperlyConfigured, self.render, template, context)

        settings.HAYSTACK_CUSTOM_HIGHLIGHTER = (
            "test_haystack.test_templatetags.BorkHighlighter"
        )

        template = """{% load highlight %}{% highlight entry with query %}"""
        context = {"entry": self.sample_entry, "query": "index"}
        self.assertEqual(
            self.render(template, context),
            "Bork!ing behavior for your model you can specify your own SearchIndex class.\nThis is useful for ensuring that future-dated or non-live content is not Bork!ed\nand searchable.\n\nEvery custom SearchIndex ",
        )

        # Restore.
        settings.HAYSTACK_CUSTOM_HIGHLIGHTER = old_custom_highlighter



from django.test import TestCase

from haystack import connections
from haystack.utils.loading import UnifiedIndex
from test_haystack.discovery.search_indexes import FooIndex

EXPECTED_INDEX_MODEL_COUNT = 6


class ManualDiscoveryTestCase(TestCase):
    def test_discovery(self):
        old_ui = connections["default"].get_unified_index()
        connections["default"]._index = UnifiedIndex()
        ui = connections["default"].get_unified_index()
        self.assertEqual(len(ui.get_indexed_models()), EXPECTED_INDEX_MODEL_COUNT)

        ui.build(indexes=[FooIndex()])

        self.assertListEqual(
            ["discovery.foo"], [str(i._meta) for i in ui.get_indexed_models()]
        )

        ui.build(indexes=[])

        self.assertListEqual([], ui.get_indexed_models())
        connections["default"]._index = old_ui


class AutomaticDiscoveryTestCase(TestCase):
    def test_discovery(self):
        old_ui = connections["default"].get_unified_index()
        connections["default"]._index = UnifiedIndex()
        ui = connections["default"].get_unified_index()
        self.assertEqual(len(ui.get_indexed_models()), EXPECTED_INDEX_MODEL_COUNT)

        # Test exclusions.
        ui.excluded_indexes = ["test_haystack.discovery.search_indexes.BarIndex"]
        ui.build()

        indexed_model_names = [str(i._meta) for i in ui.get_indexed_models()]
        self.assertIn("multipleindex.foo", indexed_model_names)
        self.assertIn("multipleindex.bar", indexed_model_names)
        self.assertNotIn("discovery.bar", indexed_model_names)

        ui.excluded_indexes = [
            "test_haystack.discovery.search_indexes.BarIndex",
            "test_haystack.discovery.search_indexes.FooIndex",
        ]
        ui.build()

        indexed_model_names = [str(i._meta) for i in ui.get_indexed_models()]
        self.assertIn("multipleindex.foo", indexed_model_names)
        self.assertIn("multipleindex.bar", indexed_model_names)
        self.assertListEqual(
            [], [i for i in indexed_model_names if i.startswith("discovery")]
        )
        connections["default"]._index = old_ui



from django.test.client import RequestFactory
from django.test.testcases import TestCase

from haystack.forms import ModelSearchForm
from haystack.generic_views import SearchView


class GenericSearchViewsTestCase(TestCase):
    """Test case for the generic search views."""

    def setUp(self):
        super().setUp()
        self.query = "haystack"
        self.request = self.get_request(url="/some/random/url?q={0}".format(self.query))

    def test_get_form_kwargs(self):
        """Test getting the search view form kwargs."""
        v = SearchView()
        v.request = self.request

        form_kwargs = v.get_form_kwargs()
        self.assertEqual(form_kwargs.get("data").get("q"), self.query)
        self.assertEqual(form_kwargs.get("initial"), {})
        self.assertTrue("searchqueryset" in form_kwargs)
        self.assertTrue("load_all" in form_kwargs)

    def test_search_view_response(self):
        """Test the generic SearchView response."""
        response = SearchView.as_view()(request=self.request)

        context = response.context_data
        self.assertEqual(context["query"], self.query)
        self.assertEqual(context.get("view").__class__, SearchView)
        self.assertEqual(context.get("form").__class__, ModelSearchForm)
        self.assertIn("page_obj", context)
        self.assertNotIn("page", context)

    def test_search_view_form_valid(self):
        """Test the generic SearchView form is valid."""
        v = SearchView()
        v.kwargs = {}
        v.request = self.request

        form = v.get_form(v.get_form_class())
        response = v.form_valid(form)
        context = response.context_data

        self.assertEqual(context["query"], self.query)

    def test_search_view_form_invalid(self):
        """Test the generic SearchView form is invalid."""
        v = SearchView()
        v.kwargs = {}
        v.request = self.request

        form = v.get_form(v.get_form_class())
        response = v.form_invalid(form)
        context = response.context_data

        self.assertTrue("query" not in context)

    def get_request(self, url, method="get", data=None, **kwargs):
        """Gets the request object for the view.

        :param url: a mock url to use for the request
        :param method: the http method to use for the request ('get', 'post',
            etc).
        """
        factory = RequestFactory()
        factory_func = getattr(factory, method)

        request = factory_func(url, data=data or {}, **kwargs)
        return request



from django.apps import apps
from django.db import models

from haystack import connection_router, connections
from haystack.exceptions import NotHandled
from haystack.query import SearchQuerySet
from haystack.signals import BaseSignalProcessor, RealtimeSignalProcessor

from ..whoosh_tests.testcases import WhooshTestCase
from .models import Bar, Foo
from .search_indexes import BarIndex, FooIndex


class MultipleIndexTestCase(WhooshTestCase):
    def setUp(self):
        super().setUp()

        self.ui = connections["solr"].get_unified_index()
        self.fi = self.ui.get_index(Foo)
        self.bi = self.ui.get_index(Bar)
        self.solr_backend = connections["solr"].get_backend()
        self.whoosh_backend = connections["whoosh"].get_backend()
        self.filtered_whoosh_backend = connections["filtered_whoosh"].get_backend()

        Foo.objects.bulk_create(
            [
                Foo(title="Haystack test", body="foo 1"),
                Foo(title="Another Haystack test", body="foo 2"),
            ]
        )

        Bar.objects.bulk_create(
            [
                Bar(author="Haystack test", content="bar 1"),
                Bar(author="Another Haystack test", content="bar 2"),
                Bar(author="Yet another Haystack test", content="bar 3"),
            ]
        )

        self.fi.reindex(using="solr")
        self.fi.reindex(using="whoosh")
        self.bi.reindex(using="solr")

    def tearDown(self):
        self.fi.clear(using="solr")
        self.bi.clear(using="solr")
        super().tearDown()

    def test_index_update_object_using(self):
        results = self.solr_backend.search("foo")
        self.assertEqual(results["hits"], 2)
        results = self.whoosh_backend.search("foo")
        self.assertEqual(results["hits"], 2)

        foo_3 = Foo.objects.create(title="Whee another Haystack test", body="foo 3")

        self.fi.update_object(foo_3, using="solr")
        results = self.solr_backend.search("foo")
        self.assertEqual(results["hits"], 3)
        results = self.whoosh_backend.search("foo")
        self.assertEqual(results["hits"], 2)

        self.fi.update_object(foo_3, using="whoosh")
        results = self.solr_backend.search("foo")
        self.assertEqual(results["hits"], 3)
        results = self.whoosh_backend.search("foo")
        self.assertEqual(results["hits"], 3)

    def test_index_remove_object_using(self):
        results = self.solr_backend.search("foo")
        self.assertEqual(results["hits"], 2)
        results = self.whoosh_backend.search("foo")
        self.assertEqual(results["hits"], 2)

        foo_1 = Foo.objects.get(pk=1)

        self.fi.remove_object(foo_1, using="solr")
        results = self.solr_backend.search("foo")
        self.assertEqual(results["hits"], 1)
        results = self.whoosh_backend.search("foo")
        self.assertEqual(results["hits"], 2)

        self.fi.remove_object(foo_1, using="whoosh")
        results = self.solr_backend.search("foo")
        self.assertEqual(results["hits"], 1)
        results = self.whoosh_backend.search("foo")
        self.assertEqual(results["hits"], 1)

    def test_index_clear_using(self):
        results = self.solr_backend.search("foo")
        self.assertEqual(results["hits"], 2)
        results = self.whoosh_backend.search("foo")
        self.assertEqual(results["hits"], 2)

        self.fi.clear(using="solr")
        results = self.solr_backend.search("foo")
        self.assertEqual(results["hits"], 0)
        results = self.whoosh_backend.search("foo")
        self.assertEqual(results["hits"], 2)

        self.fi.clear(using="whoosh")
        results = self.solr_backend.search("foo")
        self.assertEqual(results["hits"], 0)
        results = self.whoosh_backend.search("foo")
        self.assertEqual(results["hits"], 0)

    def test_index_update_using(self):
        self.fi.clear(using="solr")
        self.fi.clear(using="whoosh")
        self.bi.clear(using="solr")
        self.bi.clear(using="whoosh")

        results = self.solr_backend.search("foo")
        self.assertEqual(results["hits"], 0)
        results = self.whoosh_backend.search("foo")
        self.assertEqual(results["hits"], 0)

        self.fi.update(using="solr")
        results = self.solr_backend.search("foo")
        self.assertEqual(results["hits"], 2)
        results = self.whoosh_backend.search("foo")
        self.assertEqual(results["hits"], 0)

        self.fi.update(using="whoosh")
        results = self.solr_backend.search("foo")
        self.assertEqual(results["hits"], 2)
        results = self.whoosh_backend.search("foo")
        self.assertEqual(results["hits"], 2)

    def test_searchqueryset_using(self):
        # Using the default.
        sqs = SearchQuerySet("solr")
        self.assertEqual(sqs.count(), 5)
        self.assertEqual(sqs.models(Foo).count(), 2)
        self.assertEqual(sqs.models(Bar).count(), 3)

        self.assertEqual(sqs.using("solr").count(), 5)
        self.assertEqual(sqs.using("solr").models(Foo).count(), 2)
        self.assertEqual(sqs.using("solr").models(Bar).count(), 3)

        self.assertEqual(sqs.using("whoosh").count(), 2)
        self.assertEqual(sqs.using("whoosh").models(Foo).count(), 2)
        self.assertEqual(sqs.using("whoosh").models(Bar).count(), 0)

    def test_searchquery_using(self):
        sq = connections["solr"].get_query()

        # Using the default.
        self.assertEqual(sq.get_count(), 5)

        # "Swap" to the default.
        sq = sq.using("solr")
        self.assertEqual(sq.get_count(), 5)

        # Swap the ``SearchQuery`` used.
        sq = sq.using("whoosh")
        self.assertEqual(sq.get_count(), 2)

    def test_excluded_indexes(self):
        wui = connections["filtered_whoosh"].get_unified_index()
        self.assertTrue(any(isinstance(i, FooIndex) for i in wui.collect_indexes()))
        self.assertFalse(any(isinstance(i, BarIndex) for i in wui.collect_indexes()))

        # Shouldn't error.
        wui.get_index(Foo)

        # Should error, since it's not present.
        self.assertRaises(NotHandled, wui.get_index, Bar)

    def test_filtered_index_update(self):
        for i in ("whoosh", "filtered_whoosh"):
            self.fi.clear(using=i)
            self.fi.update(using=i)

        results = self.whoosh_backend.search("foo")
        self.assertEqual(results["hits"], 2)

        results = self.filtered_whoosh_backend.search("foo")
        self.assertEqual(
            results["hits"], 1, "Filtered backend should only contain one record"
        )


class TestSignalProcessor(BaseSignalProcessor):
    def setup(self):
        self.setup_ran = True
        super().setup()

    def teardown(self):
        self.teardown_ran = True
        super().teardown()


class SignalProcessorTestCase(WhooshTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        config = apps.get_app_config("haystack")
        cls._old_sp = config.signal_processor
        config.signal_processor = RealtimeSignalProcessor(
            connections, connection_router
        )

    @classmethod
    def tearDown(cls):
        config = apps.get_app_config("haystack")
        config.signal_processor.teardown()
        config.signal_processor = cls._old_sp
        super().tearDown()

    def setUp(self):
        super().setUp()

        # Blatantly wrong data, just for assertion purposes.
        self.fake_connections = {}
        self.fake_router = []

        self.ui = connections["solr"].get_unified_index()
        self.fi = self.ui.get_index(Foo)
        self.bi = self.ui.get_index(Bar)
        self.solr_backend = connections["solr"].get_backend()
        self.whoosh_backend = connections["whoosh"].get_backend()

        self.foo_1 = Foo.objects.create(title="Haystack test", body="foo 1")
        self.foo_2 = Foo.objects.create(title="Another Haystack test", body="foo 2")
        self.bar_1 = Bar.objects.create(author="Haystack test", content="bar 1")
        self.bar_2 = Bar.objects.create(author="Another Haystack test", content="bar 2")
        self.bar_3 = Bar.objects.create(
            author="Yet another Haystack test", content="bar 3"
        )

        self.fi.reindex(using="solr")
        self.fi.reindex(using="whoosh")
        self.bi.reindex(using="solr")

    def tearDown(self):
        self.fi.clear(using="solr")
        self.bi.clear(using="solr")
        super().tearDown()

    def test_init(self):
        tsp = TestSignalProcessor(self.fake_connections, self.fake_router)
        self.assertEqual(tsp.connections, self.fake_connections)
        self.assertEqual(tsp.connection_router, self.fake_router)
        # We fake some side-effects to make sure it ran.
        self.assertTrue(tsp.setup_ran)

        bsp = BaseSignalProcessor(self.fake_connections, self.fake_router)
        self.assertFalse(getattr(bsp, "setup_ran", False))

    def test_setup(self):
        tsp = TestSignalProcessor(self.fake_connections, self.fake_router)
        tsp.setup()
        self.assertTrue(tsp.setup_ran)

    def test_teardown(self):
        tsp = TestSignalProcessor(self.fake_connections, self.fake_router)
        tsp.teardown()
        self.assertTrue(tsp.teardown_ran)

    def test_handle_save(self):
        # Because the code here is pretty leaky (abstraction-wise), we'll test
        # the actual setup.
        # First, ensure the signal is setup.
        self.assertEqual(len(models.signals.post_save.receivers), 1)

        # Second, check the existing search data.
        sqs = SearchQuerySet("solr")
        self.assertEqual(sqs.using("solr").count(), 5)
        self.assertEqual(sqs.using("solr").models(Foo).count(), 2)
        self.assertEqual(sqs.using("solr").models(Bar).count(), 3)
        self.assertEqual(sqs.using("whoosh").count(), 2)
        self.assertEqual(sqs.using("whoosh").models(Foo).count(), 2)

        self.assertEqual(
            sqs.using("solr").models(Foo).order_by("django_id")[0].text, "foo 1"
        )
        self.assertEqual(
            sqs.using("whoosh").models(Foo).order_by("django_id")[0].text, "foo 1"
        )

        # Third, save the model, which should fire the signal & index the
        # new data.
        self.foo_1.body = "A different body"
        self.foo_1.save()

        # Fourth, check the search data for the updated data, making sure counts
        # haven't changed.
        sqs = SearchQuerySet("solr")
        self.assertEqual(sqs.using("solr").count(), 5)
        self.assertEqual(sqs.using("solr").models(Foo).count(), 2)
        self.assertEqual(sqs.using("solr").models(Bar).count(), 3)
        self.assertEqual(sqs.using("whoosh").count(), 2)
        self.assertEqual(sqs.using("whoosh").models(Foo).count(), 2)

        self.assertEqual(
            sqs.using("solr").models(Foo).order_by("django_id")[0].text,
            "A different body",
        )
        self.assertEqual(
            sqs.using("whoosh").models(Foo).order_by("django_id")[0].text, "foo 1"
        )

    def test_handle_delete(self):
        # Because the code here is pretty leaky (abstraction-wise), we'll test
        # the actual setup.
        # First, ensure the signal is setup.
        self.assertEqual(len(models.signals.post_delete.receivers), 1)

        # Second, check the existing search data.
        sqs = SearchQuerySet("solr")
        self.assertEqual(sqs.using("solr").count(), 5)
        self.assertEqual(sqs.using("solr").models(Foo).count(), 2)
        self.assertEqual(sqs.using("solr").models(Bar).count(), 3)
        self.assertEqual(sqs.using("whoosh").count(), 2)
        self.assertEqual(sqs.using("whoosh").models(Foo).count(), 2)

        self.assertEqual(
            sqs.using("solr").models(Foo).order_by("django_id")[0].text, "foo 1"
        )
        self.assertEqual(
            sqs.using("whoosh").models(Foo).order_by("django_id")[0].text, "foo 1"
        )

        # Third, delete the model, which should fire the signal & remove the
        # record from the index.
        self.foo_1.delete()

        # Fourth, check the search data for the now-removed data, making sure counts
        # have changed correctly.
        sqs = SearchQuerySet("solr")
        self.assertEqual(sqs.using("solr").count(), 4)
        self.assertEqual(sqs.using("solr").models(Foo).count(), 1)
        self.assertEqual(sqs.using("solr").models(Bar).count(), 3)
        self.assertEqual(sqs.using("whoosh").count(), 2)
        self.assertEqual(sqs.using("whoosh").models(Foo).count(), 2)

        self.assertEqual(
            sqs.using("solr").models(Foo).order_by("django_id")[0].text, "foo 2"
        )
        self.assertEqual(
            sqs.using("whoosh").models(Foo).order_by("django_id")[0].text, "foo 1"
        )



from unittest.mock import call, patch

from django.conf import settings
from django.core.management import call_command
from django.test import TestCase

__all__ = ["CoreManagementCommandsTestCase"]


class CoreManagementCommandsTestCase(TestCase):
    @patch("haystack.management.commands.update_index.Command.update_backend")
    def test_update_index_default_using(self, m):
        """update_index uses default index when --using is not present"""
        call_command("update_index")
        for k in settings.HAYSTACK_CONNECTIONS:
            self.assertTrue(call("core", k) in m.call_args_list)

    @patch("haystack.management.commands.update_index.Command.update_backend")
    def test_update_index_using(self, m):
        """update_index only applies to indexes specified with --using"""
        call_command("update_index", verbosity=0, using=["eng", "fra"])
        m.assert_any_call("core", "eng")
        m.assert_any_call("core", "fra")
        self.assertTrue(
            call("core", "default") not in m.call_args_list,
            "update_index should have been restricted to the index specified with --using",
        )

    @patch("haystack.loading.ConnectionHandler.__getitem__")
    def test_clear_index_default_using(self, m):
        """clear_index uses all keys when --using is not present"""
        call_command("clear_index", verbosity=0, interactive=False)
        self.assertEqual(len(settings.HAYSTACK_CONNECTIONS), m.call_count)
        for k in settings.HAYSTACK_CONNECTIONS:
            self.assertTrue(call(k) in m.call_args_list)

    @patch("haystack.loading.ConnectionHandler.__getitem__")
    def test_clear_index_using(self, m):
        """clear_index only applies to indexes specified with --using"""

        call_command("clear_index", verbosity=0, interactive=False, using=["eng"])
        m.assert_called_with("eng")
        self.assertTrue(
            m.return_value.get_backend.called, "backend.clear() should be called"
        )
        self.assertTrue(
            call("default") not in m.call_args_list,
            "clear_index should have been restricted to the index specified with --using",
        )

    @patch("haystack.loading.ConnectionHandler.__getitem__")
    @patch("haystack.management.commands.update_index.Command.update_backend")
    def test_rebuild_index_default_using(self, m1, m2):
        """rebuild_index uses default index when --using is not present"""

        call_command("rebuild_index", verbosity=0, interactive=False)
        self.assertEqual(len(settings.HAYSTACK_CONNECTIONS), m2.call_count)
        for k in settings.HAYSTACK_CONNECTIONS:
            self.assertTrue(call(k) in m2.call_args_list)
        m1.assert_any_call("core", "default")
        m1.assert_any_call("core", "whoosh")

    @patch("haystack.loading.ConnectionHandler.__getitem__")
    @patch("haystack.management.commands.update_index.Command.update_backend")
    def test_rebuild_index_using(self, m1, m2):
        """rebuild_index passes --using to clear_index and update_index"""

        call_command("rebuild_index", verbosity=0, interactive=False, using=["eng"])
        m2.assert_called_with("eng")
        m1.assert_any_call("core", "eng")

    @patch("haystack.management.commands.update_index.Command.handle", return_value="")
    @patch("haystack.management.commands.clear_index.Command.handle", return_value="")
    def test_rebuild_index(self, mock_handle_clear, mock_handle_update):
        call_command("rebuild_index", interactive=False)

        self.assertTrue(mock_handle_clear.called)
        self.assertTrue(mock_handle_update.called)

    @patch("haystack.management.commands.update_index.Command.handle", return_value="")
    @patch("haystack.management.commands.clear_index.Command.handle", return_value="")
    def test_rebuild_index_nocommit(self, *mocks):
        call_command("rebuild_index", interactive=False, commit=False)

        for m in mocks:
            self.assertEqual(m.call_count, 1)

            args, kwargs = m.call_args

            self.assertIn("commit", kwargs)
            self.assertEqual(False, kwargs["commit"])

    @patch("haystack.management.commands.clear_index.Command.handle", return_value="")
    @patch("haystack.management.commands.update_index.Command.handle", return_value="")
    def test_rebuild_index_nocommit_two(self, update_mock, clear_mock):
        """
        Confirm that command-line option parsing produces the same results as using call_command() directly,
        mostly as a sanity check for the logic in rebuild_index which combines the option_lists for its
        component commands.
        """
        from haystack.management.commands.rebuild_index import Command

        Command().run_from_argv(
            ["django-admin.py", "rebuild_index", "--noinput", "--nocommit"]
        )

        for m in (clear_mock, update_mock):
            self.assertEqual(m.call_count, 1)

            args, kwargs = m.call_args

            self.assertIn("commit", kwargs)
            self.assertEqual(False, kwargs["commit"])

        args, kwargs = clear_mock.call_args

        self.assertIn("interactive", kwargs)
        self.assertIs(kwargs["interactive"], False)



import os
import shutil

from django.conf import settings
from django.test import TestCase


class WhooshTestCase(TestCase):
    fixtures = ["base_data"]

    @classmethod
    def setUpClass(cls):
        for name, conn_settings in settings.HAYSTACK_CONNECTIONS.items():
            if (
                conn_settings["ENGINE"]
                != "haystack.backends.whoosh_backend.WhooshEngine"
            ):
                continue

            if "STORAGE" in conn_settings and conn_settings["STORAGE"] != "file":
                continue

            # Start clean
            if os.path.exists(conn_settings["PATH"]):
                shutil.rmtree(conn_settings["PATH"])

            from haystack import connections

            connections[name].get_backend().setup()

        super().setUpClass()

    @classmethod
    def tearDownClass(cls):
        for conn in settings.HAYSTACK_CONNECTIONS.values():
            if conn["ENGINE"] != "haystack.backends.whoosh_backend.WhooshEngine":
                continue

            if "STORAGE" in conn and conn["STORAGE"] != "file":
                continue

            # Start clean
            if os.path.exists(conn["PATH"]):
                shutil.rmtree(conn["PATH"])

        super().tearDownClass()



import datetime
import queue
import time
from threading import Thread

from django.test import TestCase

from haystack import connections, indexes
from haystack.exceptions import SearchFieldError
from haystack.utils.loading import UnifiedIndex
from test_haystack.core.models import (
    AFifthMockModel,
    AnotherMockModel,
    AThirdMockModel,
    ManyToManyLeftSideModel,
    ManyToManyRightSideModel,
    MockModel,
)


class BadSearchIndex1(indexes.SearchIndex, indexes.Indexable):
    author = indexes.CharField(model_attr="author")
    pub_date = indexes.DateTimeField(model_attr="pub_date")

    def get_model(self):
        return MockModel


class BadSearchIndex2(indexes.SearchIndex, indexes.Indexable):
    text = indexes.CharField(document=True, use_template=True)
    content2 = indexes.CharField(document=True, use_template=True)
    author = indexes.CharField(model_attr="author")
    pub_date = indexes.DateTimeField(model_attr="pub_date")

    def get_model(self):
        return MockModel


class GoodMockSearchIndex(indexes.SearchIndex, indexes.Indexable):
    text = indexes.CharField(document=True, use_template=True)
    author = indexes.CharField(model_attr="author")
    pub_date = indexes.DateTimeField(model_attr="pub_date")
    extra = indexes.CharField(indexed=False, use_template=True)

    def get_model(self):
        return MockModel


# For testing inheritance...
class AltGoodMockSearchIndex(GoodMockSearchIndex, indexes.Indexable):
    additional = indexes.CharField(model_attr="author")

    def get_model(self):
        return MockModel


class GoodCustomMockSearchIndex(indexes.SearchIndex, indexes.Indexable):
    text = indexes.CharField(document=True, use_template=True)
    author = indexes.CharField(model_attr="author", faceted=True)
    pub_date = indexes.DateTimeField(model_attr="pub_date", faceted=True)
    extra = indexes.CharField(indexed=False, use_template=True)
    hello = indexes.CharField(model_attr="hello")

    def prepare(self, obj):
        super().prepare(obj)
        self.prepared_data["whee"] = "Custom preparation."
        return self.prepared_data

    def prepare_author(self, obj):
        return "Hi, I'm %s" % self.prepared_data["author"]

    def load_all_queryset(self):
        return self.get_model()._default_manager.filter(id__gt=1)

    def get_model(self):
        return MockModel

    def index_queryset(self, using=None):
        return MockModel.objects.all()

    def read_queryset(self, using=None):
        return MockModel.objects.filter(author__in=["daniel1", "daniel3"])

    def build_queryset(self, start_date=None, end_date=None):
        return MockModel.objects.filter(author__in=["daniel1", "daniel3"])


class GoodNullableMockSearchIndex(indexes.SearchIndex, indexes.Indexable):
    text = indexes.CharField(document=True, use_template=True)
    author = indexes.CharField(model_attr="author", null=True, faceted=True)

    def get_model(self):
        return MockModel


class GoodOverriddenFieldNameMockSearchIndex(indexes.SearchIndex, indexes.Indexable):
    text = indexes.CharField(
        document=True, use_template=True, index_fieldname="more_content"
    )
    author = indexes.CharField(model_attr="author", index_fieldname="name_s")
    hello = indexes.CharField(model_attr="hello")

    def get_model(self):
        return MockModel


class GoodFacetedMockSearchIndex(indexes.SearchIndex, indexes.Indexable):
    text = indexes.CharField(document=True, use_template=True)
    author = indexes.CharField(model_attr="author")
    author_foo = indexes.FacetCharField(facet_for="author")
    pub_date = indexes.DateTimeField(model_attr="pub_date")
    pub_date_exact = indexes.FacetDateTimeField(facet_for="pub_date")

    def get_model(self):
        return MockModel

    def prepare_author(self, obj):
        return "Hi, I'm %s" % self.prepared_data["author"]

    def prepare_pub_date_exact(self, obj):
        return "2010-10-26T01:54:32"


class MROFieldsSearchIndexA(indexes.SearchIndex, indexes.Indexable):
    text = indexes.CharField(document=True, model_attr="test_a")

    def get_model(self):
        return MockModel


class MROFieldsSearchIndexB(indexes.SearchIndex, indexes.Indexable):
    text = indexes.CharField(document=True, model_attr="test_b")

    def get_model(self):
        return MockModel


class MROFieldsSearchChild(MROFieldsSearchIndexA, MROFieldsSearchIndexB):
    pass


class ModelWithManyToManyFieldAndAttributeLookupSearchIndex(
    indexes.SearchIndex, indexes.Indexable
):
    text = indexes.CharField(document=True)
    related_models = indexes.MultiValueField(model_attr="related_models__name")

    def get_model(self):
        return ManyToManyLeftSideModel


class SearchIndexTestCase(TestCase):
    fixtures = ["base_data"]

    def setUp(self):
        super().setUp()
        self.sb = connections["default"].get_backend()
        self.mi = GoodMockSearchIndex()
        self.cmi = GoodCustomMockSearchIndex()
        self.cnmi = GoodNullableMockSearchIndex()
        self.gfmsi = GoodFacetedMockSearchIndex()

        # Fake the unified index.
        self.old_unified_index = connections["default"]._index
        self.ui = UnifiedIndex()
        self.ui.build(indexes=[self.mi])
        connections["default"]._index = self.ui

        self.sample_docs = {
            "core.mockmodel.1": {
                "text": "Indexed!\n1\n",
                "django_id": "1",
                "django_ct": "core.mockmodel",
                "extra": "Stored!\n1",
                "author": "daniel1",
                "pub_date": datetime.datetime(2009, 3, 17, 6, 0),
                "id": "core.mockmodel.1",
            },
            "core.mockmodel.2": {
                "text": "Indexed!\n2\n",
                "django_id": "2",
                "django_ct": "core.mockmodel",
                "extra": "Stored!\n2",
                "author": "daniel2",
                "pub_date": datetime.datetime(2009, 3, 17, 7, 0),
                "id": "core.mockmodel.2",
            },
            "core.mockmodel.3": {
                "text": "Indexed!\n3\n",
                "django_id": "3",
                "django_ct": "core.mockmodel",
                "extra": "Stored!\n3",
                "author": "daniel3",
                "pub_date": datetime.datetime(2009, 3, 17, 8, 0),
                "id": "core.mockmodel.3",
            },
        }

    def tearDown(self):
        connections["default"]._index = self.old_unified_index
        super().tearDown()

    def test_no_contentfield_present(self):
        self.assertRaises(SearchFieldError, BadSearchIndex1)

    def test_too_many_contentfields_present(self):
        self.assertRaises(SearchFieldError, BadSearchIndex2)

    def test_contentfield_present(self):
        try:
            mi = GoodMockSearchIndex()
        except:
            self.fail()

    def test_proper_fields(self):
        self.assertEqual(len(self.mi.fields), 4)
        self.assertTrue("text" in self.mi.fields)
        self.assertTrue(isinstance(self.mi.fields["text"], indexes.CharField))
        self.assertTrue("author" in self.mi.fields)
        self.assertTrue(isinstance(self.mi.fields["author"], indexes.CharField))
        self.assertTrue("pub_date" in self.mi.fields)
        self.assertTrue(isinstance(self.mi.fields["pub_date"], indexes.DateTimeField))
        self.assertTrue("extra" in self.mi.fields)
        self.assertTrue(isinstance(self.mi.fields["extra"], indexes.CharField))

        self.assertEqual(len(self.cmi.fields), 7)
        self.assertTrue("text" in self.cmi.fields)
        self.assertTrue(isinstance(self.cmi.fields["text"], indexes.CharField))
        self.assertTrue("author" in self.cmi.fields)
        self.assertTrue(isinstance(self.cmi.fields["author"], indexes.CharField))
        self.assertTrue("author_exact" in self.cmi.fields)
        self.assertTrue(
            isinstance(self.cmi.fields["author_exact"], indexes.FacetCharField)
        )
        self.assertTrue("pub_date" in self.cmi.fields)
        self.assertTrue(isinstance(self.cmi.fields["pub_date"], indexes.DateTimeField))
        self.assertTrue("pub_date_exact" in self.cmi.fields)
        self.assertTrue(
            isinstance(self.cmi.fields["pub_date_exact"], indexes.FacetDateTimeField)
        )
        self.assertTrue("extra" in self.cmi.fields)
        self.assertTrue(isinstance(self.cmi.fields["extra"], indexes.CharField))
        self.assertTrue("hello" in self.cmi.fields)
        self.assertTrue(isinstance(self.cmi.fields["extra"], indexes.CharField))

    def test_index_queryset(self):
        self.assertEqual(self.cmi.index_queryset().count(), 3)

    def test_read_queryset(self):
        self.assertEqual(self.cmi.read_queryset().count(), 2)

    def test_build_queryset(self):
        # The custom SearchIndex.build_queryset returns the same records as
        # the read_queryset
        self.assertEqual(self.cmi.build_queryset().count(), 2)

        # Store a reference to the original method
        old_guf = self.mi.__class__.get_updated_field

        self.mi.__class__.get_updated_field = lambda self: "pub_date"

        # With an updated field, we should get have filtered results
        sd = datetime.datetime(2009, 3, 17, 7, 0)
        self.assertEqual(len(self.mi.build_queryset(start_date=sd)), 2)

        ed = datetime.datetime(2009, 3, 17, 7, 59)
        self.assertEqual(len(self.mi.build_queryset(end_date=ed)), 2)

        sd = datetime.datetime(2009, 3, 17, 6, 0)
        ed = datetime.datetime(2009, 3, 17, 6, 59)
        self.assertEqual(len(self.mi.build_queryset(start_date=sd, end_date=ed)), 1)

        # Remove the updated field for the next test
        del self.mi.__class__.get_updated_field

        # The default should return all 3 even if we specify a start date
        # because there is no updated field specified
        self.assertEqual(len(self.mi.build_queryset(start_date=sd)), 3)

        # Restore the original attribute
        self.mi.__class__.get_updated_field = old_guf

    def test_prepare(self):
        mock = MockModel()
        mock.pk = 20
        mock.author = "daniel%s" % mock.id
        mock.pub_date = datetime.datetime(2009, 1, 31, 4, 19, 0)

        self.assertEqual(len(self.mi.prepare(mock)), 7)
        self.assertEqual(
            sorted(self.mi.prepare(mock).keys()),
            ["author", "django_ct", "django_id", "extra", "id", "pub_date", "text"],
        )

    def test_custom_prepare(self):
        mock = MockModel()
        mock.pk = 20
        mock.author = "daniel%s" % mock.id
        mock.pub_date = datetime.datetime(2009, 1, 31, 4, 19, 0)

        self.assertEqual(len(self.cmi.prepare(mock)), 11)
        self.assertEqual(
            sorted(self.cmi.prepare(mock).keys()),
            [
                "author",
                "author_exact",
                "django_ct",
                "django_id",
                "extra",
                "hello",
                "id",
                "pub_date",
                "pub_date_exact",
                "text",
                "whee",
            ],
        )

        self.assertEqual(len(self.cmi.full_prepare(mock)), 11)
        self.assertEqual(
            sorted(self.cmi.full_prepare(mock).keys()),
            [
                "author",
                "author_exact",
                "django_ct",
                "django_id",
                "extra",
                "hello",
                "id",
                "pub_date",
                "pub_date_exact",
                "text",
                "whee",
            ],
        )

    def test_thread_safety(self):
        # This is a regression. ``SearchIndex`` used to write to
        # ``self.prepared_data``, which would leak between threads if things
        # went too fast.
        exceptions = []

        def threaded_prepare(index_queue, index, model):
            try:
                index.queue = index_queue
                prepped = index.prepare(model)
            except Exception as e:
                exceptions.append(e)
                raise

        class ThreadedSearchIndex(GoodMockSearchIndex):
            def prepare_author(self, obj):
                if obj.pk == 20:
                    time.sleep(0.1)
                else:
                    time.sleep(0.5)

                index_queue.put(self.prepared_data["author"])
                return self.prepared_data["author"]

        tmi = ThreadedSearchIndex()
        index_queue = queue.Queue()
        mock_1 = MockModel()
        mock_1.pk = 20
        mock_1.author = "foo"
        mock_1.pub_date = datetime.datetime(2009, 1, 31, 4, 19, 0)
        mock_2 = MockModel()
        mock_2.pk = 21
        mock_2.author = "daniel%s" % mock_2.id
        mock_2.pub_date = datetime.datetime(2009, 1, 31, 4, 19, 0)

        th1 = Thread(target=threaded_prepare, args=(index_queue, tmi, mock_1))
        th2 = Thread(target=threaded_prepare, args=(index_queue, tmi, mock_2))

        th1.start()
        th2.start()
        th1.join()
        th2.join()

        mock_1_result = index_queue.get()
        mock_2_result = index_queue.get()
        self.assertEqual(mock_1_result, "foo")
        self.assertEqual(mock_2_result, "daniel21")

    def test_custom_prepare_author(self):
        mock = MockModel()
        mock.pk = 20
        mock.author = "daniel%s" % mock.id
        mock.pub_date = datetime.datetime(2009, 1, 31, 4, 19, 0)

        self.assertEqual(len(self.cmi.prepare(mock)), 11)
        self.assertEqual(
            sorted(self.cmi.prepare(mock).keys()),
            [
                "author",
                "author_exact",
                "django_ct",
                "django_id",
                "extra",
                "hello",
                "id",
                "pub_date",
                "pub_date_exact",
                "text",
                "whee",
            ],
        )

        self.assertEqual(len(self.cmi.full_prepare(mock)), 11)
        self.assertEqual(
            sorted(self.cmi.full_prepare(mock).keys()),
            [
                "author",
                "author_exact",
                "django_ct",
                "django_id",
                "extra",
                "hello",
                "id",
                "pub_date",
                "pub_date_exact",
                "text",
                "whee",
            ],
        )
        self.assertEqual(self.cmi.prepared_data["author"], "Hi, I'm daniel20")
        self.assertEqual(self.cmi.prepared_data["author_exact"], "Hi, I'm daniel20")

    def test_custom_model_attr(self):
        mock = MockModel()
        mock.pk = 20
        mock.author = "daniel%s" % mock.id
        mock.pub_date = datetime.datetime(2009, 1, 31, 4, 19, 0)

        self.assertEqual(len(self.cmi.prepare(mock)), 11)
        self.assertEqual(
            sorted(self.cmi.prepare(mock).keys()),
            [
                "author",
                "author_exact",
                "django_ct",
                "django_id",
                "extra",
                "hello",
                "id",
                "pub_date",
                "pub_date_exact",
                "text",
                "whee",
            ],
        )

        self.assertEqual(len(self.cmi.full_prepare(mock)), 11)
        self.assertEqual(
            sorted(self.cmi.full_prepare(mock).keys()),
            [
                "author",
                "author_exact",
                "django_ct",
                "django_id",
                "extra",
                "hello",
                "id",
                "pub_date",
                "pub_date_exact",
                "text",
                "whee",
            ],
        )
        self.assertEqual(self.cmi.prepared_data["hello"], "World!")

    def test_custom_index_fieldname(self):
        mock = MockModel()
        mock.pk = 20
        mock.author = "daniel%s" % mock.id
        mock.pub_date = datetime.datetime(2009, 1, 31, 4, 19, 0)

        cofnmi = GoodOverriddenFieldNameMockSearchIndex()
        self.assertEqual(len(cofnmi.prepare(mock)), 6)
        self.assertEqual(
            sorted(cofnmi.prepare(mock).keys()),
            ["django_ct", "django_id", "hello", "id", "more_content", "name_s"],
        )
        self.assertEqual(cofnmi.prepared_data["name_s"], "daniel20")
        self.assertEqual(cofnmi.get_content_field(), "more_content")

    def test_get_content_field(self):
        self.assertEqual(self.mi.get_content_field(), "text")

    def test_update(self):
        self.sb.clear()
        self.assertEqual(self.sb.search("*")["hits"], 0)
        self.mi.update()
        self.assertEqual(self.sb.search("*")["hits"], 3)
        self.sb.clear()

    def test_update_object(self):
        self.sb.clear()
        self.assertEqual(self.sb.search("*")["hits"], 0)

        mock = MockModel()
        mock.pk = 20
        mock.author = "daniel%s" % mock.id
        mock.pub_date = datetime.datetime(2009, 1, 31, 4, 19, 0)

        self.mi.update_object(mock)
        self.assertEqual(
            [(res.content_type(), res.pk) for res in self.sb.search("*")["results"]],
            [("core.mockmodel", "20")],
        )
        self.sb.clear()

    def test_remove_object(self):
        self.mi.update()
        self.assertEqual(self.sb.search("*")["hits"], 3)

        mock = MockModel()
        mock.pk = 20
        mock.author = "daniel%s" % mock.id
        mock.pub_date = datetime.datetime(2009, 1, 31, 4, 19, 0)

        self.mi.update_object(mock)
        self.assertEqual(self.sb.search("*")["hits"], 4)

        self.mi.remove_object(mock)
        self.assertEqual(
            [(res.content_type(), res.pk) for res in self.sb.search("*")["results"]],
            [("core.mockmodel", "1"), ("core.mockmodel", "2"), ("core.mockmodel", "3")],
        )

        # Put it back so we can test passing kwargs.
        mock = MockModel()
        mock.pk = 20
        mock.author = "daniel%s" % mock.id
        mock.pub_date = datetime.datetime(2009, 1, 31, 4, 19, 0)

        self.mi.update_object(mock)
        self.assertEqual(self.sb.search("*")["hits"], 4)

        self.mi.remove_object(mock, commit=False)
        self.assertEqual(
            [(res.content_type(), res.pk) for res in self.sb.search("*")["results"]],
            [
                ("core.mockmodel", "1"),
                ("core.mockmodel", "2"),
                ("core.mockmodel", "3"),
                ("core.mockmodel", "20"),
            ],
        )

        self.sb.clear()

    def test_clear(self):
        self.mi.update()
        self.assertGreater(self.sb.search("*")["hits"], 0)

        self.mi.clear()
        self.assertEqual(self.sb.search("*")["hits"], 0)

    def test_reindex(self):
        self.mi.reindex()
        self.assertEqual(
            [(res.content_type(), res.pk) for res in self.sb.search("*")["results"]],
            [("core.mockmodel", "1"), ("core.mockmodel", "2"), ("core.mockmodel", "3")],
        )
        self.sb.clear()

    def test_inheritance(self):
        try:
            agmi = AltGoodMockSearchIndex()
        except:
            self.fail()

        self.assertEqual(len(agmi.fields), 5)
        self.assertTrue("text" in agmi.fields)
        self.assertTrue(isinstance(agmi.fields["text"], indexes.CharField))
        self.assertTrue("author" in agmi.fields)
        self.assertTrue(isinstance(agmi.fields["author"], indexes.CharField))
        self.assertTrue("pub_date" in agmi.fields)
        self.assertTrue(isinstance(agmi.fields["pub_date"], indexes.DateTimeField))
        self.assertTrue("extra" in agmi.fields)
        self.assertTrue(isinstance(agmi.fields["extra"], indexes.CharField))
        self.assertTrue("additional" in agmi.fields)
        self.assertTrue(isinstance(agmi.fields["additional"], indexes.CharField))

    def test_proper_field_resolution(self):
        mrofsc = MROFieldsSearchChild()
        mock = MockModel()
        mock.pk = 20
        mock.author = "daniel%s" % mock.id
        mock.pub_date = datetime.datetime(2009, 1, 31, 4, 19, 0)
        mock.test_a = "This is A"
        mock.test_b = "This is B"

        self.assertEqual(len(mrofsc.fields), 1)
        prepped_data = mrofsc.prepare(mock)
        self.assertEqual(len(prepped_data), 4)
        self.assertEqual(prepped_data["text"], "This is A")

    def test_load_all_queryset(self):
        self.assertEqual([obj.id for obj in self.cmi.load_all_queryset()], [2, 3])

    def test_nullable(self):
        mock = MockModel()
        mock.pk = 20
        mock.author = None
        mock.pub_date = datetime.datetime(2009, 1, 31, 4, 19, 0)

        prepared_data = self.cnmi.prepare(mock)
        self.assertEqual(len(prepared_data), 6)
        self.assertEqual(
            sorted(prepared_data.keys()),
            ["author", "author_exact", "django_ct", "django_id", "id", "text"],
        )

        prepared_data = self.cnmi.full_prepare(mock)
        self.assertEqual(len(prepared_data), 4)
        self.assertEqual(
            sorted(prepared_data.keys()), ["django_ct", "django_id", "id", "text"]
        )

    def test_custom_facet_fields(self):
        mock = MockModel()
        mock.pk = 20
        mock.author = "daniel"
        mock.pub_date = datetime.datetime(2009, 1, 31, 4, 19, 0)

        prepared_data = self.gfmsi.prepare(mock)
        self.assertEqual(len(prepared_data), 8)
        self.assertEqual(
            sorted(prepared_data.keys()),
            [
                "author",
                "author_foo",
                "django_ct",
                "django_id",
                "id",
                "pub_date",
                "pub_date_exact",
                "text",
            ],
        )

        prepared_data = self.gfmsi.full_prepare(mock)
        self.assertEqual(len(prepared_data), 8)
        self.assertEqual(
            sorted(prepared_data.keys()),
            [
                "author",
                "author_foo",
                "django_ct",
                "django_id",
                "id",
                "pub_date",
                "pub_date_exact",
                "text",
            ],
        )
        self.assertEqual(prepared_data["author_foo"], "Hi, I'm daniel")
        self.assertEqual(prepared_data["pub_date_exact"], "2010-10-26T01:54:32")


class BasicModelSearchIndex(indexes.ModelSearchIndex, indexes.Indexable):
    class Meta:
        model = MockModel


class FieldsModelSearchIndex(indexes.ModelSearchIndex, indexes.Indexable):
    class Meta:
        model = MockModel
        fields = ["author", "pub_date"]


class ExcludesModelSearchIndex(indexes.ModelSearchIndex, indexes.Indexable):
    class Meta:
        model = MockModel
        excludes = ["author", "foo"]


class FieldsWithOverrideModelSearchIndex(indexes.ModelSearchIndex, indexes.Indexable):
    foo = indexes.IntegerField(model_attr="foo")

    class Meta:
        model = MockModel
        fields = ["author", "foo"]

    def get_index_fieldname(self, f):
        if f.name == "author":
            return "author_bar"
        return f.name


class YetAnotherBasicModelSearchIndex(indexes.ModelSearchIndex, indexes.Indexable):
    text = indexes.CharField(document=True)

    class Meta:
        model = AThirdMockModel


class PolymorphicModelSearchIndex(indexes.SearchIndex, indexes.Indexable):
    text = indexes.CharField(document=True)

    author = indexes.CharField(model_attr="author")
    pub_date = indexes.DateTimeField(model_attr="pub_date")
    average_delay = indexes.FloatField(null=True)

    def get_model(self):
        return AnotherMockModel

    def prepare(self, obj):
        self.prepared_data = super().prepare(obj)
        if isinstance(obj, AThirdMockModel):
            self.prepared_data["average_delay"] = obj.average_delay
        return self.prepared_data

    def index_queryset(self, using=None):
        return self.get_model().objects.all()


class GhettoAFifthMockModelSearchIndex(indexes.SearchIndex, indexes.Indexable):
    text = indexes.CharField(document=True)

    def get_model(self):
        return AFifthMockModel

    def index_queryset(self, using=None):
        # Index everything,
        return self.get_model().objects.complete_set()

    def read_queryset(self, using=None):
        return self.get_model().objects.all()


class ReadQuerySetTestSearchIndex(indexes.SearchIndex, indexes.Indexable):
    author = indexes.CharField(model_attr="author", document=True)

    def get_model(self):
        return AFifthMockModel

    def read_queryset(self, using=None):
        return self.get_model().objects.complete_set()


class TextReadQuerySetTestSearchIndex(indexes.SearchIndex, indexes.Indexable):
    text = indexes.CharField(model_attr="author", document=True)

    def get_model(self):
        return AFifthMockModel

    def read_queryset(self, using=None):
        return self.get_model().objects.complete_set()


class ModelWithManyToManyFieldModelSearchIndex(indexes.ModelSearchIndex):
    def get_model(self):
        return ManyToManyLeftSideModel


class ModelSearchIndexTestCase(TestCase):
    def setUp(self):
        super().setUp()
        self.sb = connections["default"].get_backend()
        self.bmsi = BasicModelSearchIndex()
        self.fmsi = FieldsModelSearchIndex()
        self.emsi = ExcludesModelSearchIndex()
        self.fwomsi = FieldsWithOverrideModelSearchIndex()
        self.yabmsi = YetAnotherBasicModelSearchIndex()
        self.m2mmsi = ModelWithManyToManyFieldModelSearchIndex()

    def test_basic(self):
        self.assertEqual(len(self.bmsi.fields), 4)
        self.assertTrue("foo" in self.bmsi.fields)
        self.assertTrue(isinstance(self.bmsi.fields["foo"], indexes.CharField))
        self.assertEqual(self.bmsi.fields["foo"].null, False)
        self.assertEqual(self.bmsi.fields["foo"].index_fieldname, "foo")
        self.assertTrue("author" in self.bmsi.fields)
        self.assertTrue(isinstance(self.bmsi.fields["author"], indexes.CharField))
        self.assertEqual(self.bmsi.fields["author"].null, False)
        self.assertTrue("pub_date" in self.bmsi.fields)
        self.assertTrue(isinstance(self.bmsi.fields["pub_date"], indexes.DateTimeField))
        self.assertTrue(
            isinstance(self.bmsi.fields["pub_date"].default, datetime.datetime)
        )
        self.assertTrue("text" in self.bmsi.fields)
        self.assertTrue(isinstance(self.bmsi.fields["text"], indexes.CharField))
        self.assertEqual(self.bmsi.fields["text"].document, True)
        self.assertEqual(self.bmsi.fields["text"].use_template, True)

    def test_fields(self):
        self.assertEqual(len(self.fmsi.fields), 3)
        self.assertTrue("author" in self.fmsi.fields)
        self.assertTrue(isinstance(self.fmsi.fields["author"], indexes.CharField))
        self.assertTrue("pub_date" in self.fmsi.fields)
        self.assertTrue(isinstance(self.fmsi.fields["pub_date"], indexes.DateTimeField))
        self.assertTrue("text" in self.fmsi.fields)
        self.assertTrue(isinstance(self.fmsi.fields["text"], indexes.CharField))

    def test_excludes(self):
        self.assertEqual(len(self.emsi.fields), 2)
        self.assertTrue("pub_date" in self.emsi.fields)
        self.assertTrue(isinstance(self.emsi.fields["pub_date"], indexes.DateTimeField))
        self.assertTrue("text" in self.emsi.fields)
        self.assertTrue(isinstance(self.emsi.fields["text"], indexes.CharField))
        self.assertNotIn("related_models", self.m2mmsi.fields)

    def test_fields_with_override(self):
        self.assertEqual(len(self.fwomsi.fields), 3)
        self.assertTrue("author" in self.fwomsi.fields)
        self.assertTrue(isinstance(self.fwomsi.fields["author"], indexes.CharField))
        self.assertTrue("foo" in self.fwomsi.fields)
        self.assertTrue(isinstance(self.fwomsi.fields["foo"], indexes.IntegerField))
        self.assertTrue("text" in self.fwomsi.fields)
        self.assertTrue(isinstance(self.fwomsi.fields["text"], indexes.CharField))

    def test_overriding_field_name_with_get_index_fieldname(self):
        self.assertTrue(self.fwomsi.fields["foo"].index_fieldname, "foo")
        self.assertTrue(self.fwomsi.fields["author"].index_fieldname, "author_bar")

    def test_float_integer_fields(self):
        self.assertEqual(len(self.yabmsi.fields), 5)
        self.assertEqual(
            sorted(self.yabmsi.fields.keys()),
            ["author", "average_delay", "pub_date", "text", "view_count"],
        )
        self.assertTrue("author" in self.yabmsi.fields)
        self.assertTrue(isinstance(self.yabmsi.fields["author"], indexes.CharField))
        self.assertEqual(self.yabmsi.fields["author"].null, False)
        self.assertTrue("pub_date" in self.yabmsi.fields)
        self.assertTrue(
            isinstance(self.yabmsi.fields["pub_date"], indexes.DateTimeField)
        )
        self.assertTrue(
            isinstance(self.yabmsi.fields["pub_date"].default, datetime.datetime)
        )
        self.assertTrue("text" in self.yabmsi.fields)
        self.assertTrue(isinstance(self.yabmsi.fields["text"], indexes.CharField))
        self.assertEqual(self.yabmsi.fields["text"].document, True)
        self.assertEqual(self.yabmsi.fields["text"].use_template, False)
        self.assertTrue("view_count" in self.yabmsi.fields)
        self.assertTrue(
            isinstance(self.yabmsi.fields["view_count"], indexes.IntegerField)
        )
        self.assertEqual(self.yabmsi.fields["view_count"].null, False)
        self.assertEqual(self.yabmsi.fields["view_count"].index_fieldname, "view_count")
        self.assertTrue("average_delay" in self.yabmsi.fields)
        self.assertTrue(
            isinstance(self.yabmsi.fields["average_delay"], indexes.FloatField)
        )
        self.assertEqual(self.yabmsi.fields["average_delay"].null, False)
        self.assertEqual(
            self.yabmsi.fields["average_delay"].index_fieldname, "average_delay"
        )


class ModelWithManyToManyFieldAndAttributeLookupSearchIndexTestCase(TestCase):
    def test_full_prepare(self):
        index = ModelWithManyToManyFieldAndAttributeLookupSearchIndex()

        left_model = ManyToManyLeftSideModel.objects.create()
        right_model_1 = ManyToManyRightSideModel.objects.create(name="Right side 1")
        right_model_2 = ManyToManyRightSideModel.objects.create()
        left_model.related_models.add(right_model_1)
        left_model.related_models.add(right_model_2)

        result = index.full_prepare(left_model)

        self.assertDictEqual(
            result,
            {
                "django_ct": "core.manytomanyleftsidemodel",
                "django_id": "1",
                "text": None,
                "id": "core.manytomanyleftsidemodel.1",
                "related_models": ["Right side 1", "Default name"],
            },
        )


class PolymorphicModelTestCase(TestCase):
    def test_prepare_with_polymorphic(self):
        index = PolymorphicModelSearchIndex()

        parent_model = AnotherMockModel()
        parent_model.author = "Paul"
        parent_model.pub_date = datetime.datetime(2018, 5, 23, 13, 57)
        parent_model.save()

        child_model = AThirdMockModel()
        child_model.author = "Paula"
        child_model.pub_date = datetime.datetime(2018, 5, 23, 13, 58)
        child_model.average_delay = 0.5
        child_model.save()

        prepared_data = index.prepare(parent_model)
        self.assertEqual(len(prepared_data), 7)
        self.assertEqual(
            sorted(prepared_data.keys()),
            [
                "author",
                "average_delay",
                "django_ct",
                "django_id",
                "id",
                "pub_date",
                "text",
            ],
        )
        self.assertEqual(prepared_data["django_ct"], "core.anothermockmodel")
        self.assertEqual(prepared_data["average_delay"], None)

        prepared_data = index.prepare(child_model)
        self.assertEqual(len(prepared_data), 7)
        self.assertEqual(
            sorted(prepared_data.keys()),
            [
                "author",
                "average_delay",
                "django_ct",
                "django_id",
                "id",
                "pub_date",
                "text",
            ],
        )
        self.assertEqual(prepared_data["django_ct"], "core.anothermockmodel")
        self.assertEqual(prepared_data["average_delay"], 0.5)
